#include "config.h"

#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include "boilerplate-log.h"
#include "boilerplate-application.h"
#include "boilerplate-window.h"
#include "boilerplate-preferences.h"

typedef struct
{
    BoilerplateWindow *window;
} BoilerplateApplicationPrivate;


G_DEFINE_TYPE_WITH_PRIVATE (BoilerplateApplication, boilerplate_application, GTK_TYPE_APPLICATION)

gboolean DEBUG = FALSE;
gboolean DEBUG_COLOR = TRUE;
gboolean VERSION = FALSE;
gboolean CSD = TRUE;

const GOptionEntry option_entries[] = {
    { "debug", 'd', 0, G_OPTION_ARG_NONE, &DEBUG, "Run in debug mode and show extra output", NULL },
    { "debug-no-color", '\0', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &DEBUG_COLOR, "Disable color when in debug mode", NULL },
    { "version", 'v', 0, G_OPTION_ARG_NONE, &VERSION, "Print release version and exit", NULL },
    { "disable-csd", 'c', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &CSD, "Disable Client Side Decorations (CSD)", NULL },
    { NULL }
};

static GActionEntry action_entries[] = {
    { "preferences", boilerplate_application_action_preferences, NULL, NULL, NULL },
    { "shortcuts", boilerplate_application_action_shortcuts, NULL, NULL, NULL },
    { "help", boilerplate_application_action_help, NULL, NULL, NULL },
    { "about", boilerplate_application_action_about, NULL, NULL, NULL },
	{ "quit", boilerplate_application_action_quit,  NULL, NULL, NULL },
};

/**
 * boilerplate_application_show_uri:
 * @user_data: The application
 * @uri: A string that contains the url to open
 *
 * A helper function for gtk_show_uri (), this function takes an application and
 * the uri to show and loads the proper screen, and checks to see if errors
 * occure. If so they are shown to the user a a dialog instead of in the
 * terminal.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_show_uri (GApplication *user_data,
                                  const gchar  *uri)
{
    GtkApplication *application;
    GtkWindow *window;
    GtkWidget *dialog;
    GError *error = NULL;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

    gtk_show_uri (window ? gtk_window_get_screen (GTK_WINDOW (window)) : gdk_screen_get_default (),
                  uri,
                  gtk_get_current_event_time (),
                  &error);

    if (error)
    {
        dialog = gtk_message_dialog_new (window ? GTK_WINDOW (window) : NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_MESSAGE_ERROR,
                                         GTK_BUTTONS_OK,
                                         "There was an error displaying help: \n%s",
                                         error->message);

        g_signal_connect (G_OBJECT (dialog), "response",
                          G_CALLBACK (gtk_widget_destroy),
                          NULL);

        gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
        gtk_widget_show (dialog);
        g_error_free (error);
    }
}

/**
 * boilerplate_application_action_preferences:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application's preferences.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_action_preferences (GSimpleAction *action,
                                            GVariant      *parameter,
                                            gpointer       user_data)
{
    GtkApplication *application;
    BoilerplateWindow *window;
    BoilerplatePreferences *preferences;

    application = GTK_APPLICATION (user_data);
    window = BOILERPLATE_WINDOW (gtk_application_get_active_window (application));

    preferences = boilerplate_preferences_new (window);

    gtk_window_present (GTK_WINDOW (preferences));
}

/**
 * boilerplate_application_action_shortcuts:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application keyboard shortcuts and provides a fallback for users
 * running <= GTK 3.20 by launching a help window linking to the keyboard
 * shortcuts page.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_action_shortcuts (GSimpleAction *action,
                                          GVariant      *parameter,
                                          gpointer       user_data)
{
    GtkApplication *application;
    BoilerplateWindow *window;
    GtkWidget *shortcuts_window = NULL;
    GtkWidget *dialog;
    GError *error = NULL;

    application = GTK_APPLICATION (user_data);
    window = BOILERPLATE_WINDOW (gtk_application_get_active_window (application));

    if (gtk_get_major_version () >= 3 && gtk_get_minor_version () >= 20 ) {
        if (shortcuts_window == NULL)
        {
            GtkBuilder *builder;

            builder = gtk_builder_new_from_resource (APPLICATION_RESOURCES "/ui/keyboard-shortcuts.ui");
            shortcuts_window = GTK_WIDGET (gtk_builder_get_object (builder, "shortcuts"));

            g_signal_connect (shortcuts_window,
                              "destroy",
                              G_CALLBACK (gtk_widget_destroyed),
                              &shortcuts_window);

            g_object_unref (builder);
        }

        if (GTK_WINDOW (window) != gtk_window_get_transient_for (GTK_WINDOW (shortcuts_window)))
        {
            gtk_window_set_transient_for (GTK_WINDOW (shortcuts_window), GTK_WINDOW (window));
        }

        gtk_widget_show_all (shortcuts_window);
        gtk_window_present (GTK_WINDOW (shortcuts_window));
    } else {
        // Fallback mode for users using a GTK version before 3.20
        boilerplate_application_show_uri (user_data,
                                          "help:" PACKAGE "?keyboard-shortcuts");
    }
}

/**
 * boilerplate_application_action_help:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application's help window.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_action_help (GSimpleAction *action,
                                     GVariant      *parameter,
                                     gpointer       user_data)
{
    boilerplate_application_show_uri (user_data,
                                      "help:" PACKAGE);
}

/**
 * boilerplate_application_action_about:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Opens the application's about dialog.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_action_about (GSimpleAction *action,
                                      GVariant      *parameter,
                                      gpointer       user_data)
{
    GtkApplication *application;
    GtkWindow *window;

    application = GTK_APPLICATION (user_data);
    window = gtk_application_get_active_window (application);

	static const char *authors[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const char *artists[] = {
		"Julian Richen <julian@richen.io>",
		NULL,
	};

	static const gchar *translators =
        "Julian Richen <julian@richen.io>";

	gtk_show_about_dialog (window,
		                   "program-name", g_get_application_name (),
		                   "version", PACKAGE_VERSION,
		                   "copyright", "Copyright \xc2\xa9 "COPYRIGHT" Julian Richen",
		                   "comments", "Short description of application.",
		                   "authors", authors,
		                   "artists", artists,
		                   "translator-credits", translators,
		                   "website-label", "Website",
		                   "website", PACKAGE_URL,
		                   "logo-icon-name", gtk_window_get_default_icon_name (),
		                   "wrap-license", TRUE,
		                   "license-type", GTK_LICENSE_GPL_3_0,
		                   NULL);
}

/**
 * boilerplate_application_action_quit:
 * @application:  The application
 * @options: GActionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Action to quit the application.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_action_quit (GSimpleAction *action,
                                     GVariant      *parameter,
                                     gpointer       user_data)
{
    g_application_quit (G_APPLICATION (user_data));
}

/**
 * boilerplate_application_init_accelerators:
 * @application:
 *
 * A helper function originally from EOG (Eye of GNOME) which inits all the
 * application level accelerators with minimal code.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_init_accelerators (GtkApplication *application)
{
	static const gchar *const accelmap[] = {
        "app.preferences", "<Primary>comma", NULL,
        "app.shortcuts", "<Primary>F1", NULL,
        "app.shortcuts", "<Primary><Shift>question", NULL,
        "app.help", "F1", NULL,
        "app.about", "F2", NULL,
        "app.quit", "<Primary>q", NULL,
		NULL
	};

	const gchar *const *it;

	for (it = accelmap; it[0]; it += g_strv_length ((gchar **)it) + 1)
    {
		gtk_application_set_accels_for_action (application, it[0], &it[1]);
	}
}

/**
 * boilerplate_application_activate:
 * @application: The application
 *
 * Application Sequence: 04
 * Activates the GtkApplicationWindow and presents the window to the user.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_activate (GApplication *application)
{
    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    BoilerplateApplicationPrivate* priv = boilerplate_application_get_instance_private (self);

    /*
     * Create GtkApplicationWindow if needed
     * If you need multiple windows open (like you have multiple instances) you
     * should convert priv->window to a GList *windows and loop the list to
     * select the window
     */
    if (priv->window == NULL) {
	    priv->window = boilerplate_window_new (BOILERPLATE_APPLICATION (application));
    }

    boilerplate_window_set_csd (BOILERPLATE_WINDOW (priv->window),
                                CSD);

	gtk_window_present (GTK_WINDOW (priv->window));
}

/**
 * boilerplate_application_startup:
 * @application: The application
 *
 * Application Sequence: 03
 * Starts many of the applications features. It also sets information used
 * throughout the application like the application:
 * - name
 * - icon
 * - appmenu
 * - menubar
 * - action entries
 * - accelerators
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_startup (GApplication *application)
{
    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);
    GtkBuilder *builder;

    g_log (NULL, G_LOG_LEVEL_DEBUG, "Starting application");

    g_application_set_resource_base_path (G_APPLICATION (self), APPLICATION_RESOURCES);

	G_APPLICATION_CLASS (boilerplate_application_parent_class)->startup (G_APPLICATION (self));

    /* Set application name */
	g_set_application_name (APPLICATION_NAME);

    /* Set default icon via name */
	gtk_window_set_default_icon_name (APPLICATION_ID);

    /* Init menus */
    /* Add action callbacks */
	g_action_map_add_action_entries (G_ACTION_MAP (application),
                                     action_entries, G_N_ELEMENTS (action_entries),
                                     application);

    /* Add acclerators (keyboard shortcuts) */
	boilerplate_application_init_accelerators (GTK_APPLICATION (application));

    /* Load glade/ui file with menus */
    builder = gtk_builder_new_from_resource (APPLICATION_RESOURCES "/ui/menus.ui");

    /* App menu */
    gtk_application_set_app_menu (GTK_APPLICATION (application),
                                  G_MENU_MODEL (gtk_builder_get_object (builder, "appmenu")));

    /* Ribbon Menu (fallback) */
    gtk_application_set_menubar (GTK_APPLICATION (application),
                                 G_MENU_MODEL (gtk_builder_get_object (builder, "menubar")));

    /* Clean-up builder */
    g_object_unref (builder);
}

/**
 * boilerplate_application_handle_local_command_line:
 * @application:  The application
 * @options: GOptionEntry arg_data
 * @user_data: User data set when the signal handler was connected
 *
 * Application Sequence: 02
 * Handles all the commandline flags like --debug, --version, -v, etc...
 *
 * Returns: -1 Continue / 0 proccess <something>
 *
 * Since: 0.1.0
 */
static gint
boilerplate_application_handle_local_command_line (GApplication *application,
                                                   GVariantDict *options,
                                                   gpointer      user_data)
{
    BoilerplateApplication *self = BOILERPLATE_APPLICATION (application);

    if (VERSION)
    {
        g_print ("%s\n", PACKAGE_VERSION);
        return 0;
    }

    return -1;
}

/*
static void
boilerplate_application_dark_theme_changed (GSettings   *settings,
                                            const char  *key,
                                            GtkSettings *gtk_settings)
{
    gboolean prefer_dark_theme = g_settings_get_boolean (settings, key);

    g_object_set (gtk_settings,
                  "gtk-application-prefer-dark-theme",
                  prefer_dark_theme,
                  NULL);
}
*/

/**
 * boilerplate_application_init:
 * @self: The application
 *
 * Initializes the application and connects the local-options/command-line
 * options to boilerplate_application_handle_local_command_line.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_init (BoilerplateApplication *self)
{
    BoilerplateApplicationPrivate *priv;

    priv = boilerplate_application_get_instance_private (self);

	g_application_add_main_option_entries (G_APPLICATION (self), option_entries);

    //TODO: Do a check to see if it exists
    self->settings = g_settings_new (APPLICATION_ID);

    g_signal_connect (self, "handle-local-options",
                      G_CALLBACK (boilerplate_application_handle_local_command_line),
                      self);


    //TODO: Dark mode setup
    /*
    GtkSettings *gtk_settings = gtk_settings_get_default ();

    boilerplate_application_dark_theme_changed (self->settings,
                                                "dark-theme",
                                                gtk_settings);
    g_signal_connect (self->settings,
                     "changed::dark-theme",
                     G_CALLBACK (boilerplate_application_dark_theme_changed),
                     gtk_settings);
    */

}

/**
 * boilerplate_application_class_init:
 * @class: The application class
 *
 * Initializes the various functions in the application.
 *
 * Returns: void
 *
 * Since: 0.1.0
 */
static void
boilerplate_application_class_init (BoilerplateApplicationClass *class)
{
	G_APPLICATION_CLASS (class)->startup = boilerplate_application_startup;
	G_APPLICATION_CLASS (class)->activate = boilerplate_application_activate;
}

/**
 * boilerplate_application_new:
 *
 * Application Sequence: 01
 * Creates a new GtkApplication.
 *
 * Returns: BoilerplateApplication -> GtkApplication
 *
 * Since: 0.1.0
 */
BoilerplateApplication *
boilerplate_application_new (void)
{
    g_log_set_handler (NULL,
                       G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
                       log_handler,
                       NULL);

    return g_object_new (BOILERPLATE_TYPE_APPLICATION,
                         "application-id", APPLICATION_ID,
                         NULL);
}
